/*
 * tape.c -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) 2005, Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>

#include "config.h"
#include "tape.h"

int32_t
tape_read (struct tapeinfo *tio, uint8_t *buf, uint32_t reclen)
{
	uint32_t	i;
	uint32_t	recl_head;
	uint32_t	recl_tail;
	uint8_t		rbuf[4];
	uint8_t		ibuf[BUFMAX];

	if (reclen > MAX_TAPE_RECLEN) {
		fprintf (stderr, "tape_read: Requested record length of %d bytes too big.  Max %d\n",
				reclen, MAX_TAPE_RECLEN);
		return(-1);
	}

	if (tio->debug)
		fprintf(stderr, "tape_read: Pre-read position: %d\n", (int) lseek (tio->fd, 0, SEEK_CUR));

	tio->reqbytes = reclen;

	i = read (tio->fd, rbuf, 4);

	if (tio->debug)
		fprintf(stderr, "tape_read: Read %d header bytes.  tio->fd=%d\n", i, tio->fd);

	if (i == 0)
		return (0);			/* Simulator end-of-tape */
	if (i < 4)
		return (-2);			/* Simulated tape format error */

	recl_head = ((uint32_t) rbuf[3] << 24 ) |
		    ((uint32_t) rbuf[2] << 16 ) |
		    ((uint32_t) rbuf[1] << 8 )  |
		     (uint32_t) rbuf[0];	/* Compute the SIMH record length from the header */

	if (tio->debug)
		fprintf(stderr, "tape_read: computed header reclen: %d\n", recl_head);

	if (recl_head > MAX_TAPE_RECLEN) {
		fprintf (stderr, "tape_read: tape header reclen of %d bytes too big.  Max %d\n",
				recl_head, MAX_TAPE_RECLEN);
		return(-3);
	}

	if (recl_head > 0) {
		if ((tio->actbytes = read (tio->fd, ibuf, recl_head)) == 0)
			return (-4);
		if (tio->debug)
			fprintf(stderr, "tape_read: data count=%d\n", tio->actbytes);

		if (tio->actbytes < 0) {
			tio->actbytes = 0;
			tio->status = TAPEIO_STATUS_ERROR;
			tio->ext_status = TAPEIO_EXT_STATUS_HDW;
			tio->hdw_error = errno;
			return (-99);
		}

		memmove (buf, ibuf, tio->actbytes);
	} else {
		if (tio->debug)
			fprintf(stderr, "tape_read: logical EOF detected.\n");
		return (0);
	}

	i = read (tio->fd, rbuf, 4);

	if (tio->debug)
		fprintf(stderr, "tape_read: Read %d trailer bytes.  tio->fd=%d\n", i, tio->fd);

	if (i < 4)
		return (-5);			/* Simulated tape format error */

	recl_tail = ((uint32_t) rbuf[3] << 24 ) |
		    ((uint32_t) rbuf[2] << 16 ) |
		    ((uint32_t) rbuf[1] << 8 )  |
		     (uint32_t) rbuf[0];	/* Compute the SIMH record length from the trailier */

	if (recl_tail > MAX_TAPE_RECLEN) {
		fprintf (stderr, "tape_read: tape trailer reclen of %d bytes too big.  Max %d\n",
				recl_tail, MAX_TAPE_RECLEN);
		return(-6);
	}

	if (tio->debug)
		fprintf(stderr, "tape_read: computed trailer reclen: %d\n", recl_tail);

	if (recl_head != recl_tail)
		return (-7);

	if (recl_head > tio->maxbsize)
		return (-8);			/* Record longer than requested */

	return (tio->actbytes);
}
