/*
 * util.c -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) 2005, Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "hptsb.h"
#include "decl.h"

/*
 * errorexit - Exit with a status code after displaying a possible error message.
 */
 
void
errorexit (int exitcode, char * format, ...)
{
        va_list args;
 
        if (format) {
                fprintf (stderr, "%s: ", progname);
                va_start (args, format);
                vfprintf (stderr, format, args);
                va_end (args);
        }
        exit (exitcode);
}
 
/*
 * eprintf -	printf(3) function to always direct the output to stderr.
 */

void
eprintf (char * format, ...)
{
	va_list args;

	va_start (args, format);
	vfprintf (stderr, format, args);
	va_end (args);
}

/*
 * lprintf -	sprintf(3) function to concatenate output onto the end of a string, rather
 *		than replacing the string.
 */

void
lprintf (char * line, char * format, ...)
{
	char 	aline[32768];
	va_list args;

	va_start (args, format);
	vsprintf (aline, format, args);
	va_end (args);
	if (strlen(line) + strlen(aline) > 32768) {
		eprintf ("\nInternal error: Adding %d bytes to lprintf string of %d bytes more than 32768 bytes.\n",
				strlen(aline), strlen(line));
	} else
		strcat (line, aline);
}

/*
 * printarrow - Print unprintable characters as ^X.
 *
 * ARGUMENTS:
 *
 * st		Destination string
 * buf		Source string
 * len		Source string length
 * method	0 = 7-bit conversion, unprintable represented as ^X
 *		1 = 8-bit no conversion, unprintable represented as \nnn
 *		(octal) and backslash (\) represented as \\.
 *
 * RETURNS:
 *		Method 0 returns the address of the Destination string.
 *		Method 1 returns the converted length of the Destination
 *		string.
 */

uint8_t *
printarrow (uint8_t *st, uint8_t * buf, uint32_t len, uint32_t method)
{
	int     i = 0;
	uint32_t	total = 0;

	st[0] = 0; 
	while (i < len && buf [i] > 0) {
		if (method) {
			if (buf [i] == '\\') {
				st[total++] = '\\';
				st[total++] = '\\';
			} else if (isprint(buf[i]))
				st[total++] = buf[i];
			else {
				sprintf (st + total, "\\%03o", buf[i]);
				total += 4;
			}
		} else {
			if ((buf[i] == 0 || buf[i] == 128))
				st[total++] = ' ';
			else if (isprint(buf [i] & 0x7f))
				st[total++] = buf[i] & 0x7f;
			else {
				st[total++] = '^';
				st[total++] = (buf[i] & 0x7f) + 64;
			}
		}
		i++;
	}
	st[total] = 0;
	if (method)
		return ((uint8_t *) total);
	return (st);
}

/*
 * HPputchar -	Put a character onto a string.
 *		If typeflag is zero, represent unprintable characters as 'N, where
 *				"n" is the decimal ASCII value of the character.
 *		If type flag is one, represent unprintable characters as \NNN, where
 *				"NNN" is the 3-digit octal representation of the character.
 *
 *		quoteflag is a pointer to an integer that keeps track of whether the
 *		printable character needs a lead-in double quote, or if an unprintable
 *		character representation (typeflag type zero) needs a closing double-
 *		quote, prior to its representation.
 *
 *		NOTE: The double-quote (") character is considered unprintable in this
 *				routine, as is the backslash (\) character..
 */

uint32_t
HPputchar (uint8_t *line, uint8_t *c, uint32_t typeflag, uint32_t *quoteflag)
{
	uint8_t		aline[255];
	aline[0] = 0;
	if (isprint(*c) && *c != '"' && *c != '\\') {
		dprintf (DEBUG_INDCONV, "%s%c", *quoteflag ? "" : "\"", *c);
		sprintf (aline, "%s%c", *quoteflag ? "" : "\"", *c);
		*quoteflag = 1;
	} else if (typeflag) {
		dprintf (DEBUG_INDCONV, "\\%d", *c);
		sprintf (aline, "\\%d", *c);
	} else {
		dprintf (DEBUG_INDCONV, "%s'%d", *quoteflag ? "\"" : "", *c);
		sprintf (aline, "%s'%d", *quoteflag ? "\"" : "", *c);
		*quoteflag = 0;
	}
	if (typeflag)
		memcpy (line, aline, strlen(aline));
	else
		strcat (line, aline);
	return (strlen(aline));
}

/*
 * ftoHP -	Convert a native floating-point number to an ASCII printable representation
 *		in the same output format used on the HP-2000/ACCESS Timeshared BASIC system.
 *
 *		UNIX printf(3) "G" format is exploited, with tweaks to match the HP TSB
 *		output format:
 *
 *		o	If the absolute value of an integer is greater than 32767, a decimal
 *			point is added.
 *		o	Irrational numbers less than one (absolute value) have no zero preceeding
 *			the devimal point.
 *		o	printf(3) "G" format represents absolute values of .0000X and .00000X in
 *			scientific notation.  HP TSB does not.
 *		o	printf(3) "G" format represents scientific notation as 1E+XX and 1E-XX.
 *			HP TSB includes a decimal point, i.e. 1.E+XX and 1.E-XX.
 */

char *
ftoHP (float fval, char * strval)
{
	char		value	[512];
	char		value2	[512];

	sprintf (value2, "%G", fval);					/* Yes */
	if ((float)((int)fval) == fval) {				/* Integer? */
		if ((abs(fval) <= 999999 && abs(fval) > 32767))		/* Add decimal point? */
			strcat (value2, ".");
		strcpy (value, value2);
	} else {							/* Irrational */
		if (value2[0] == '0' && value2[1] == '.')
			strcpy (value, value2 + 1);			/* Strip off leading zero */
		else if (value2[0] == '-' && value2[1] == '0' && value2[2] == '.')
			sprintf(value, "-%s", value2 + 2);		/* Also for negative */
		else
			strcpy (value, value2);
		/* Some hacks to get HP BASIC representations from printf(3) %G format. */
		if (!strcmp(value + 1, "E-06"))				/* Hack for .00000X */
			sprintf(value, ".00000%c", value[0]);
		if (!strcmp(value + 1, "E-05"))				/* Hack for .0000X */
			sprintf(value, ".0000%c", value[0]);
		if (!strcmp(value + 2, "E-06") && value[0] == '-')	/* Hack for -.00000X */
			sprintf(value, "-.00000%c", value[1]);
		if (!strcmp(value + 2, "E-05") && value[0] == '-')	/* Hack for -.0000X */
			sprintf(value, "-.0000%c", value[1]);
	}

	if (value[1] == 'E') {						/* Change 1E+XX to 1.E+XX */
		strcpy (value2, value);
		sprintf (value, "%c.%s", value2[0], value2 + 1);
	}
	if (value[0] == '-' && value[2] == 'E') {			/* Change 1E-XX to 1.E-XX */
		strcpy (value2, value);
		sprintf (value, "-%c.%s", value2[1], value2 + 2);
	}

	sprintf (strval, "%s", value);
	return(strval);
}

/*
 * HPftoIEEE -	Convert an HP-2000/ACCESS Timeshared BASIC 4-byte floating-point
 *		number to native (presumed IEEE) format.  This routine is written
 *		specifically with portability in mind.
 *
 *		Mantissa Sign          Mantissa                              Exponent   Exponent
 *		     |                    |                                       |        |                 
 *                   v                    v                                       v        v
 *		     - ----------------------------------------------------- ------------- -
 *		    | |                                                     |             | |
 *		     - ----------------------------------------------------- ------------- -
 *		    32 31                                                  8 7           1 0
 *
 *		The Mantissa and Exponent are 2's-compliment when negative.
 *		This routine works for both Big-Endian and Little-Endian platforms, and
 *		floating-point math is used to load the value natively rather than convert
 *		the elements directly.
 */

float
HPftoIEEE (uint32_t HPval)
{
	uint32_t	mant_sign;
	uint32_t	mant;
	uint32_t	exp;
	uint32_t	exp_sign;
	float		IEEEval;

	mant_sign = HPval & 0x80000000;
	/*
	 * If the mantissa sign is negative, the mantissa is 1's compliment (from our
	 * perspective.
	 */
	if (mant_sign)
		mant = ((~HPval) & 0x7FFFFF00) >> 8;
	else
		mant = (HPval & 0x7FFFFF00) >> 8;
	/* If the exponent sign is negative the exponent is 2's compliment. */
	exp_sign = HPval & 1;
	if (exp_sign)
		exp = (((~HPval) >> 1) & 0x7f) + 1;
	else
		exp = (HPval >> 1) & 0x7f;

	IEEEval = mant / (1.0 * (1 << 23));

	if (exp_sign)
		IEEEval /= powf(2, exp);
	else
		IEEEval *= powf(2, exp);

	if (mant_sign)
		IEEEval = -IEEEval;

	return (IEEEval);
}

/*
 * machtypename - Print the TSB System Level name.
 */

uint8_t *
machtypename (uint32_t name)
{
	switch (name) {
		case 1100:	return (MACH_TYPE_NAME_2000A);
		case 1200:	return (MACH_TYPE_NAME_2000B);
		case 1300:	return (MACH_TYPE_NAME_2000E);
		case 1500:	return (MACH_TYPE_NAME_2000C);
		case 2000:	return (MACH_TYPE_NAME_2000CP);
		case 3000:	return (MACH_TYPE_NAME_2000F);
		case 3500:	return (MACH_TYPE_NAME_2000FP);
		case 5000:	return (MACH_TYPE_NAME_ACCESSA);
		case 6000:	return (MACH_TYPE_NAME_ACCESSB);
		default:	return (MACH_TYPE_NAME_UNKNOWN);
	}
}
		
/*
 * dprintf -	Display a message on stdout, depending upon a setting and message level.
 */

void
dprintf (uint32_t level, char * format, ...)
{
	va_list args;

	if (debug & (2 << level)) {
#ifdef OLD
	if (verbose >= level) {
#endif
		va_start(args, format);
		vfprintf(stderr, format, args);
		va_end(args);
	}
}

/*
 * put_asciiline - Display a printable character (or a period (.) if unprintable)
 *		   via eprintf().
 */

void
put_asciiline(unsigned char *asciitext, int len) {
	int	j;

	for (j = 0; j < len; j++) {
		if (iscntrl((asciitext[j])&0x7f))
			eprintf(".");
		else
			eprintf("%c", (asciitext[j])&0x7f);
		if (j == 7)
			eprintf(" ");
	}
	return;
}

/*
 * hexdump -	Display a formatted hexadecimal/ASCII dump of a given block
 *		of 8-bit data.
 */

void
hexdump (uint8_t * buf, uint32_t len) {
	int		i, j;
	unsigned char	asciitext[16];

	for (i = 0; i < len; i++) {
		if (i % 16 == 0) {
			if (i != 0) {
				eprintf("  ");
				put_asciiline(asciitext, 16);
				eprintf("\n");
			}
			eprintf("%04d:  ", i);
		}
		if (i % 8 == 0)
			eprintf(" ");
		eprintf("%02X", buf[i]);
		asciitext[i % 16] = buf[i];
	}
	j = 16;
	if (i % 16 == 0)
		eprintf ("  ");
	else {
		while (i % 16) {
			j--;
			i++;
			eprintf("  ");
			if (i % 8 == 0)
				eprintf (" ");
		}
		eprintf(" ");
	}
	put_asciiline(asciitext, j);
	eprintf("\n");
	return;
}

/*
 * day2date - Display on stderr the ascii representation of the date
 *		specified by year, hours and tenths.
 */

void
day2date (uint32_t year, uint32_t hours, uint32_t tenths)
{
	uint32_t	months[12] = {744,672,744,720,744,720,744,744,720,744,720,744};
	uint8_t *	monthname[12] =
				{"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
	uint32_t	month = 0, day = 0, hour = 0, minutes, seconds;

	year += 1900;
	if (!(year % 4) && (year % 100 || !(year % 400)))
		months[1] += 24;

	for (month = 1; month < 13; month++) {
		if (hours < months[month - 1])
			break;
		hours -= months[month - 1];
	}

	day = hours / 24;
	hour = hours % 24;
	minutes = (65536 - tenths) / 600;
	seconds = ((65536 - tenths) % 600) / 10;
	tenths = ((65536 - tenths) % 600) % 10;

#ifdef ALTERNATEDATE
	eprintf ("%s %d, %d ; %04d/%02d/%02d  %02d:%02d:%02d.%d", 
			monthname[month - 1], day, year, year, month, day, hour, minutes, seconds, tenths);
#else	/* ALTERNATEDATE */
	eprintf ("%s %d, %d %02d:%02d:%02d.%d", 
			monthname[month - 1], day, year, hour, minutes, seconds, tenths);
#endif	/* ALTERNATEDATE */
}

/*
 * add_line - Add a string 'data' of 'len' bytes onto the struct
 *		'*ltab'.
 */

void
add_line (struct line_table **ltab, uint32_t len, uint8_t *data)
{
	(*ltab)->len = len;
	(*ltab)->data = calloc (1, (*ltab)->len);
	dprintf (DEBUG_FILES, "add_line: Adding %d bytes to loc %d\n",
			(*ltab)->len, (*ltab)->data);
	memcpy((*ltab)->data, data, (*ltab)->len);
	(*ltab)->next = calloc (1, sizeof (struct line_table));
	*ltab = (*ltab)->next;
	(*ltab)->len = 0;
	(*ltab)->data = NULL;
	(*ltab)->next = NULL;
}

/*
 * wildcmp - Perform a comparison between 'wildcard' and 'target'.
 *		wildcard may contain the '*' and '?' wildcards.
 */

uint32_t
wildcmp (uint8_t * wildcard, uint8_t * target)
{
	uint32_t	i, j;

	j = 0;
	for (i = 0; i < strlen(wildcard) && j < strlen(target); i++) {
		if (wildcard[i] == '*') {
			i++;
			if (i == strlen(wildcard))
				return (0);
			while (j < strlen(target)) {
				if (wildcard[i] == target[j])
					break;
				j++;
			}
			if (j == strlen(target))
				return (1);
		} else if (wildcard[i] != '?' && wildcard[i] != target[j])
			break;
		j++;
	}
	if (i < strlen(wildcard) || j < strlen(target))
		return(1);
	return(0);
}

/*
 * check_entrylist - Compare 'item' to list 'list', possibly including
 *			wildcards, and return 1 for a match.
 */

uint32_t
check_entrylist (uint8_t *item, uint8_t *list, uint32_t nlen, uint32_t nlist)
{
	uint32_t	i;

	for (i = 0; i < nlist; i++) {
		if (!wildcmp (list, item))
			return(1);
		dprintf (DEBUG_DETOKEN, "check_entrylist: Checking %s against %s\n",
				item, list);
		list += nlen;
	}
	return 0;
}
